/*
 * Copyright (c) 2018, salesforce.com, inc.
 * All rights reserved.
 * Licensed under the BSD 3-Clause license.
 * For full license text, see LICENSE.txt file in the repo root or https://opensource.org/licenses/BSD-3-Clause
 */
import fs from 'node:fs/promises';
import { ensureString } from '@salesforce/ts-types';
import chalk from 'chalk';
import { BaseDitamap } from './ditamap/base-ditamap.js';
import { CLIReference } from './ditamap/cli-reference.js';
import { Command } from './ditamap/command.js';
import { TopicCommands } from './ditamap/topic-commands.js';
import { TopicDitamap } from './ditamap/topic-ditamap.js';
import { events, punctuate } from './utils.js';
import { HelpReference } from './ditamap/help-reference.js';
function emitNoTopicMetadataWarning(topic) {
    events.emit('warning', `No metadata for topic ${chalk.bold(topic)}. That topic owner must add topic metadata in the oclif section in the package.json file within their plugin.`);
}
export class Docs {
    outputDir;
    hidden;
    topicMeta;
    cliMeta;
    constructor(outputDir, hidden, topicMeta, cliMeta) {
        this.outputDir = outputDir;
        this.hidden = hidden;
        this.topicMeta = topicMeta;
        this.cliMeta = cliMeta;
    }
    async build(commands) {
        // Create if doesn't exist
        await fs.mkdir(this.outputDir, { recursive: true });
        await this.populateTemplate(commands);
    }
    async populateTopic(topic, subtopics) {
        const topicMeta = this.topicMeta.get(topic);
        if (!topicMeta) {
            throw new Error(`No topic meta for ${topic} - add this topic to the oclif section of the package.json.`);
        }
        let description = topicMeta.description;
        if (!description && !topicMeta.external) {
            // TODO: check why the same property is used again when it is already used above
            description = punctuate(topicMeta.description);
            if (!description) {
                events.emit('warning', `No description for topic ${chalk.bold(topic)}. Skipping until topic owner adds topic metadata in the oclif section in the package.json file within their plugin.`);
                return [];
            }
        }
        const subTopicNames = [];
        const commandIds = [];
        for (const [subtopic, classes] of subtopics.entries()) {
            try {
                // const subTopicsMeta = topicMeta.subtopics;
                // if (!subTopicsMeta?.get(subtopic)) {
                //   emitNoTopicMetadataWarning(`${topic}:${subtopic}`);
                //   continue;
                // }
                subTopicNames.push(subtopic);
                // Commands within the sub topic
                for (const command of classes) {
                    const fullTopic = ensureString(command.id).replace(/:\w+$/, '');
                    const commandsInFullTopic = classes.filter((cmd) => ensureString(cmd.id).startsWith(fullTopic));
                    const commandMeta = this.resolveCommandMeta(ensureString(command.id), command, commandsInFullTopic.length);
                    // eslint-disable-next-line no-await-in-loop
                    await this.populateCommand(topic, subtopic, command, commandMeta);
                    commandIds.push(command.id);
                }
            }
            catch (error) {
                const err = error instanceof Error ? error : typeof error === 'string' ? new Error(error) : new Error('Unknown error');
                if (err.name === 'UnexpectedValueTypeError') {
                    emitNoTopicMetadataWarning(`${topic}:${subtopic}`);
                }
                else {
                    events.emit('warning', `Can't create topic for ${topic}:${subtopic}: ${err.message}\n`);
                }
            }
        }
        // The topic ditamap with all of the subtopic links.
        events.emit('subtopics', topic, subTopicNames);
        await new TopicCommands(topic, topicMeta).write();
        await new TopicDitamap(topic, commandIds).write();
        return subTopicNames;
    }
    /**
     * Group all commands by the top level topic and then subtopic. e.g. force, analytics, evergreen, etc
     * then org, apex, etc within the force namespace.
     *
     * @param commands - The entire set of command data.
     * @returns The commands grouped by topics/subtopic/commands.
     */
    groupTopicsAndSubtopics(commands) {
        // const topLevelTopics: Dictionary<Dictionary<CommandClass | CommandClass[]>> = {};
        const topLevelTopics = new Map();
        for (const command of commands) {
            if (command.hidden && !this.hidden) {
                continue;
            }
            const commandParts = ensureString(command.id).split(':');
            const topLevelTopic = commandParts[0];
            const plugin = command.plugin;
            if (plugin) {
                // Also include the namespace on the commands so we don't need to do the split at other times in the code.
                command.topic = topLevelTopic;
                const existingTopicsForTopLevel = topLevelTopics.get(topLevelTopic) ?? new Map();
                if (commandParts.length === 1) {
                    // This is a top-level topic that is also a command
                    const existingTarget = existingTopicsForTopLevel.get(commandParts[0]) ?? [];
                    existingTopicsForTopLevel.set(commandParts[0], [...existingTarget, command]);
                }
                else if (commandParts.length === 2) {
                    // This is a command directly under the top-level topic
                    const existingTarget = existingTopicsForTopLevel.get(commandParts[1]) ?? [];
                    existingTopicsForTopLevel.set(commandParts[1], [...existingTarget, command]);
                }
                else {
                    const subtopic = commandParts[1];
                    try {
                        const topicMeta = this.topicMeta.get(topLevelTopic);
                        const subTopicsMeta = topicMeta?.subtopics?.get(subtopic);
                        if (subTopicsMeta?.hidden && !this.hidden) {
                            continue;
                        }
                    }
                    catch (e) {
                        // It means no meta so it isn't hidden, although it should always fail before here with no meta found
                    }
                    command.subtopic = subtopic;
                    const subtopicCommands = existingTopicsForTopLevel.get(subtopic) ?? [];
                    existingTopicsForTopLevel.set(subtopic, [...subtopicCommands, command]);
                }
                topLevelTopics.set(topLevelTopic, existingTopicsForTopLevel);
            }
        }
        return topLevelTopics;
    }
    async populateTemplate(commands) {
        const topicsAndSubtopics = this.groupTopicsAndSubtopics(commands);
        await new CLIReference().write();
        await new HelpReference().write();
        // Generate one base file with all top-level topics.
        await new BaseDitamap(Array.from(topicsAndSubtopics.keys())).write();
        for (const [topic, subtopics] of topicsAndSubtopics.entries()) {
            events.emit('topic', { topic });
            // eslint-disable-next-line no-await-in-loop
            await this.populateTopic(topic, subtopics);
        }
    }
    resolveCommandMeta(commandId, command, commandsInTopic) {
        const commandMeta = Object.assign({}, this.cliMeta);
        // Remove top level topic, since the topic meta is already for that topic
        const commandParts = commandId.split(':');
        let part;
        try {
            let currentMeta;
            for (part of commandParts) {
                if (currentMeta) {
                    const subtopics = currentMeta.subtopics;
                    currentMeta = subtopics?.get(part);
                }
                else {
                    currentMeta = this.topicMeta.get(part);
                }
                // Collect all tiers of the meta, so the command will also pick up the topic state (isPilot, etc) if applicable
                Object.assign({}, commandMeta, currentMeta);
            }
        }
        catch (error) {
            // @ts-expect-error: part may be undefined
            if (commandId.endsWith(part)) {
                // This means there wasn't meta information going all the way down to the command, which is ok.
                return commandMeta;
            }
            else if (commandsInTopic !== 1) {
                events.emit('warning', `subtopic "${part}" meta not found for command ${commandId}`);
            }
            else if (!commandMeta.description) {
                commandMeta.description = command.description;
                commandMeta.longDescription = (command.longDescription ? command.longDescription : punctuate(command.description));
            }
        }
        return commandMeta;
    }
    async populateCommand(topic, subtopic, command, commandMeta) {
        // If it is a hidden command - abort
        if (command.hidden && !this.hidden) {
            return '';
        }
        const commandDitamap = new Command(topic, subtopic, command, commandMeta);
        await commandDitamap.write();
        return commandDitamap.getFilename();
    }
}
//# sourceMappingURL=docs.js.map